/***********************************************************************
**
**   Sclassify Ver 1.0 by Gangman Yi
**
***********************************************************************/

#include<cstdlib>
#include<time.h>
#include<iostream>
#include<fstream>
#include<string>
#include<map>
#include<stack>
#include<stdio.h>
#include<math.h>
#include<sstream>
#include<cstdio>

using namespace std;

#define VERSION		"1.0"
#define EVALUE_VAL	3							// type definition evalue_val type is defined by number 3

static char  UNKNOWN_FAMILY[]="UNKNOWN";
static char  OUTPUT_SEPARATOR = '\t';
static float MAX_EVALUE = 745.0;
static float MIN_EVALUE = -0.01;

map<int, map<int, float> > distance_matrix_score;				// Distance matrix of blast score 
map<int, map<int, float> > distance_matrix_evalue;				// Distance matrix of blast e-value by -log(e-value)
map<int, string> protein_id2name, protein_id2family;				// protein idx returns its protein name, protein_id2family returns its family name
map<string, int> protein_name2id;						// protein name returns its protein idx. 
map<string, map<int, int> > family_id_list;					// family name returns its protein idxs.   family_id_list[family_name][#] = [protein id]
map<int, map<int, int> > unknown_protein_pair;					// unknown_protein_pair[unknown protein_id][#] = known protein_id
map<int, map<int, int> > unknown_to_unknown_pair;				// unknown_to_unknown_pair[unknown protein_id][#] = unknown protein_id
map<int, map<int, float> > test2test_distance_matrix_evalue;			// Distance matrix of blast e-value by -log(e-value)

ofstream output_file_fp;
int total_num_known_proteins = 0;


void OPTIONS(string err1="", string err2="")
{ 
	ostringstream OPTIONS;
	
	if( err1 != "" ) {
		cerr << "\nError! " << "[1m " << err1 << "[0m ";
		if( err2 != "" ) cerr << " " << err2;
		cerr << endl;
	}
	
	OPTIONS << "\
	\n   ::[1m sclassify[0m v" << VERSION << " ::					\n\n   \
	-c : classified protein family list							\n \
	-u : unclassified protein list								\n \
	-e : e-value cutoff [default=1e-10]							\n \
	-p : e-value list between unclassified to classified protein 				\n \
	-n : e-value list between unclassified to unclassified protein 				\n \
	-o : output file name									\n\n\n \
	[-c] A file that lists the name of each protein in existing families along with 	\n \
	     the name of its family in a two-column tab-separated format			\n \
	     (example file:pfam.list). 								\n\n \
	[-u] A file that lists the name of each unclassified protein in a one-column 		\n \
	     format (example file: test.list). 							\n\n \
	[-p] A file that lists the e-values between each unclassified protein and each 		\n \
	     protein in existing families in a three-column tab-separated format that 		\n \
	     gives the name of an unclassified protein, the name of a protein in an 		\n \
	     existing family, and the e-value between them.  There is no need to have 		\n \
	     an e-value for each pair if some of them are missing.  The file can also 		\n \
	     be left blank (example file: test_pfam.score). 					\n\n \
	[-n] A file that lists the e-values between each pair of unclassified proteins 		\n \
	     in a three-column tab-separated format that gives the names of two 		\n \
	     unclassified proteins and the e-value between them.  There is no need to 		\n \
	     have an e-value for each pair if some of them are missing.  The file can 		\n \
	     also be left blank (example file: test_test.score). 				\n\n";

	cout << OPTIONS.str();
	exit(0);
}


void CHECK_START(int argc, char* argv[], string *training_family_file_name, string *testing_family_file_name, string *testing_pair_file_name, string *test_test_pair_file_name, float *eval_cutoff, string *out_file_name)
{	

	string classified_list = "", unclassified_list = "", classified_pair = "", unclassified_pair = "", result_file_name = "";
	double e_value = 1e-10;

	if( argc < 6) OPTIONS();
	
	for( int i=1; i<argc; i++)
	{
		if( i % 2 != 0 ){
			if( argv[i][0] != '-' ) OPTIONS(argv[i]);
			if( argc <= i+1 ) OPTIONS();
			
			switch(argv[i][1])
			{
				case 'c' : classified_list.assign(argv[i+1]);
				           break;
				case 'u' : unclassified_list.assign(argv[i+1]);
				           break;
				case 'p' : classified_pair.assign(argv[i+1]);
				           break;
				case 'n' : unclassified_pair.assign(argv[i+1]);
				           break;
				case 'o' : result_file_name.assign(argv[i+1]);
				           break;
				case 'e' : if( sscanf( argv[i+1], "%lf", &e_value ) != 1 ) OPTIONS(argv[i], argv[i+1]);
				            break;

				default  : OPTIONS(argv[i]);
			}
			i++;
		}
	}

	if( classified_list   == "" ) OPTIONS("A file 'classified protein family list' is required.");
	if( unclassified_list == "" ) OPTIONS("A file 'unclassified protein list' is required.");
	if( classified_pair   == "" ) OPTIONS("A file 'e-value list between unclassified to classified protein' is required.");
	if( result_file_name  == "" ) OPTIONS("An output file is required.");

	//if( unclassified_pair == "" ) OPTIONS("A file 'e-value list between unclassified to unclassified protein' is required.");

	*training_family_file_name = classified_list;
	*testing_family_file_name  = unclassified_list;
	*testing_pair_file_name    = classified_pair;
	*test_test_pair_file_name  = unclassified_pair;
	*eval_cutoff               = e_value;
	*out_file_name             = result_file_name;
	
	cout << "\n::[1m sclassify[0m v" << VERSION << " ::\n\n";
	
}

/*
**  Checking starting and ending time
*/
void ELAPSED_TIME(time_t start_time, time_t end_time)
{
	long int dt,hr,min,sec;
	double elapsed_time;

	elapsed_time = difftime(end_time,start_time);
	
	dt = (long int)(elapsed_time / (60*60*24));
	hr = (long int)((elapsed_time - dt*60*60*24) / (60*60));
	min = (long int)((elapsed_time - dt*60*60*24 - hr*60*60)/60);
	sec = (int)elapsed_time%60;

	//cout << "\nFINISHING TIME : " << asctime (localtime(&end_time));
	printf("\n[Time elapsed: %d days %d hr %d min %d sec]\n\n",(int)dt, (int)hr, (int)min, (int)sec );
}


void COPY_BUFFER(string *dest, string *src, int length)
{
	for( int i = 0; i < length; i++ ){
		dest[i] = src[i];
	}
}


/*
**  Splitting string by some chars
*/
int SPLIT(string input, string *&result, char SPLITWord)
{
        if( result != NULL ){
                return -1;
        } else {
                result = new string[1];
        }

        string tmpString;
        int input_length = input.length(), num_words = 0;

        if( input_length ){
                for( int i=0; i<input_length; i++ ){
                        if( (input[i] == SPLITWord) && (!tmpString.empty()) ){
                                if(num_words > 0){
                                        string *tmp_StringRes = new string[num_words];
                                        COPY_BUFFER(tmp_StringRes, result, num_words);
                                        delete[] result;
                                        result = new string[num_words+1];
                                        COPY_BUFFER(result, tmp_StringRes, num_words);
                                        result[num_words] = tmpString;
                                        delete[] tmp_StringRes;
                                        num_words++;
                                } else {
                                        result[num_words] = tmpString;
                                        num_words++;
                                }
                                tmpString.erase();
                                i++;
                        }
                        tmpString += input[i];
                }
                if( tmpString.length() > 0 ){
                        if( num_words > 0 ) {
                                string *tmp_StringRes = new string[num_words];
                                COPY_BUFFER(tmp_StringRes, result, num_words);
                                delete[] result;
                                result = new string[num_words+1];
                                COPY_BUFFER(result, tmp_StringRes, num_words);
                                result[num_words] = tmpString;
                                delete[] tmp_StringRes;
                                num_words++;
                        } else {
                                result[num_words] = tmpString;
                                num_words++;
                        }
                }
        }
        return num_words;
}


/*
**  Converting string to long double
*/
long double STRTOLD(string s)
{
	string *value = NULL, tmpS;
	long double first, second =  1.0, pow_value;
	int cnt, i, maxLen = 50, p;

	if( s[0] == 'e' ) s = '1'+s;

	tmpS.assign(s);
	cnt = SPLIT(tmpS, value, 'e');

	if( cnt < 1 ){
		//cout << "Error : splitting string (" << s << ")." << endl;
		return 0;
	}else if( cnt == 1 ){
		return (long double)(atof(value[0].c_str()));
	}

	first = (long double)(atof(value[0].c_str()));
	p = atoi(value[1].c_str());

	if( p > 0 ){
		pow_value = pow(10,maxLen);
	}else{
		pow_value = pow(10,-maxLen);
	}

	for(i = 0; i < abs(p)/maxLen; i++)
		second *= pow_value;
	second *= (long double)(pow(10, p % maxLen));
	return first*second;
}


/*
**  Converting string to int
*/
int STRTOINT(const string &str) 
{
	stringstream ss(str);
	int n;
	
	ss >> n;
	return n;
}


float GET_DIST_MAT(int i, int j, int type)			// index i and j is the protein id index
{
	map<int, map<int, float> >::iterator iter;
	int idx[2];
	
	if( i > j ) {
		idx[0] = i;
		idx[1] = j;
	} else if( i < j ) {
		idx[0] = j;
		idx[1] = i;
	} else if( i == j) {
		switch( type ){
			case EVALUE_VAL : return MAX_EVALUE;
		} 		
	}	
	
	if( type == EVALUE_VAL ) {
		if( distance_matrix_evalue.find(idx[0]) != distance_matrix_evalue.end() ){
			if( distance_matrix_evalue[idx[0]].find(idx[1]) != distance_matrix_evalue[idx[0]].end() ){ 
				return distance_matrix_evalue[idx[0]][idx[1]];
			}
		}
		return MIN_EVALUE;
	}
	
	cerr << endl << "Mismatched value type[" << type << "] in GET_DIST_MAT " << endl;
	exit(0);

}


void PUT_DIST_MAT(int i, int j, float value, int type)		// index i and j is the protein id index
{
	int idx[2];
	
	if( i > j ) {
		idx[0] = i;
		idx[1] = j;
	} else if( i < j ) {
		idx[0] = j;
		idx[1] = i;
	} else {
		return;
	}

	if( type == EVALUE_VAL ) {
		distance_matrix_evalue[idx[0]][idx[1]]= value;
	} else {
		cerr << endl << "Mismatched value type[" << type << "]  in PUT_DIST_MAT" << endl;
		exit(0);
	}		
}


float GET_TEST2TEST_DIST_MAT(int i, int j, int type)			// index i and j is the protein id index
{
	map<int, map<int, float> >::iterator iter;
	int idx[2];
	
	if( i > j ) {
		idx[0] = i;
		idx[1] = j;
	} else if( i < j ) {
		idx[0] = j;
		idx[1] = i;
	} else if( i == j) {
		switch( type ){
			case EVALUE_VAL : return MAX_EVALUE;
		} 		
	}	
	
	if( type == EVALUE_VAL ) {
		if( test2test_distance_matrix_evalue.find(idx[0]) != test2test_distance_matrix_evalue.end() ){
			if( test2test_distance_matrix_evalue[idx[0]].find(idx[1]) != test2test_distance_matrix_evalue[idx[0]].end() ){ 
				return test2test_distance_matrix_evalue[idx[0]][idx[1]];
			}
		}
		return MIN_EVALUE;
	}
	
	cerr << endl << "Mismatched value type[" << type << "] in GET_DIST_MAT " << endl;
	exit(0);

}


void PUT_TEST2TEST_DIST_MAT(int i, int j, float value, int type)	// index i and j is the protein id index
{
	int idx[2];
	
	if( i > j ) {
		idx[0] = i;
		idx[1] = j;
	} else if( i < j ) {
		idx[0] = j;
		idx[1] = i;
	} else {
		return;
	}

	if( type == EVALUE_VAL ) {
		test2test_distance_matrix_evalue[idx[0]][idx[1]]= value;
	} else {
		cerr << endl << "Mismatched value type[" << type << "]  in PUT_DIST_MAT" << endl;
		exit(0);
	}		
}


int PROTEIN_NAME_EXISTS(string name)
{
	map<string, int>::iterator iter;
	
	if( protein_name2id.find(name) != protein_name2id.end() ){
		return 1; 
	} else {
		return 0;
	}
}


/*
**  Read a protein id list and protein pairs
*/
void READ_TRAINING_SET_FILES(string family_id_file_name)
{
	FILE *fp_family_id = fopen(family_id_file_name.c_str(), "r");
	char ch, tmp_char[2], tmp_ID[20];
	string tmp_str[2], tmp_pair[3];
	int i=0, j=0, tmp_i=0, tmp_score, tmp_idx;
	float score_value, evalue;
	
	if( fp_family_id == NULL ){
		cout << "Error!  Unable to open the classified protein family list file [" << family_id_file_name << "] " << endl<< endl<< endl;
		exit(1);
	} 
	
	cout << "=> Reading the classified protein family list file..." << endl;
	cout.flush();
	
	tmp_str[0].assign("");
	tmp_str[1].assign("");

	while( (ch = getc(fp_family_id)) != EOF){
		sprintf(tmp_char,"%c",ch);
		if( ch == '\n' || ch == ' ' ) {	
			if( tmp_str[0] == ""  && tmp_str[1] == "" ) continue;
			// Save information from the input file. 
			// 1st column = tmp_str[0], 2nd column = tmp_str[1] , i is index
			protein_id2name[i] = tmp_str[0];
			protein_name2id[tmp_str[0]] = i;
			protein_id2family[i] = tmp_str[1];
			family_id_list[tmp_str[1]][family_id_list[tmp_str[1]].size()] = i;
			
			// cout << "printe idx " << i << " = " <<tmp_str[0] << "\t" << tmp_str[1]<< endl;
						
			i++;
			total_num_known_proteins++;
			tmp_str[0].assign("");
			tmp_str[1].assign("");
			tmp_i = 0;
		} else if( ch == '\t'){
			tmp_i = 1;
		} else {
			tmp_str[tmp_i].append(tmp_char);
		}
	}
	fclose(fp_family_id);
	
	cout << "   : Total number of families = " << family_id_list.size() << endl;
	cout << "   : Total number of proteins = " << i << endl << endl;
	cout.flush();

}


void READ_TESTING_SET_FILES(string family_id_file_name, string protein_pair_file_name, float seq_overlapping_cutoff)
{
	ifstream fp_family_id(family_id_file_name.c_str());
	ifstream fp_pair(protein_pair_file_name.c_str());
	
	char ch, tmp_char[2], tmp_ID[20];
	string tmp_str[2], tmp_pair[4], lineStr, lineStr_arr;
	int i, j=0, tmp_i=0, tmp_score, tmp_idx, starting_idx = 0, arr_n_idx, check_flag = 0;
	float score_value, evalue;
	
	if( fp_family_id == NULL ){
		cout << "Error!  Unable to open the unclassified protein list file [" << family_id_file_name << "] " << endl<< endl<< endl;
		exit(1);
	} 
	if( fp_pair.fail() ){
		cout << "Error!  Unable to open the unclassified-classified protein score file [" << protein_pair_file_name << "] " << endl<< endl<< endl;
		exit(1);
	} 
	
	cout << "=> Reading the unclassified protein list file..." << endl;
	cout.flush();
	
	i = protein_id2name.size();
	
	while( getline(fp_family_id, lineStr) ){
		tmp_str[0].assign("");

		istringstream iss;
		iss.str(lineStr);
		for(arr_n_idx = 0; arr_n_idx < 1; arr_n_idx++){
			iss >> lineStr_arr;
			tmp_str[arr_n_idx].assign(lineStr_arr);
		}

		protein_id2name[i] = tmp_str[0];
		protein_name2id[tmp_str[0]] = i;
		protein_id2family[i] = UNKNOWN_FAMILY;
		family_id_list[UNKNOWN_FAMILY][family_id_list[UNKNOWN_FAMILY].size()] = i;
		
		// cout << starting_idx << "\t[" << tmp_str[0] << "]\t" << i << endl;
		starting_idx++;
		i++;
	}
	fp_family_id.close();

	cout << "   : Total number of unclassified proteins = " << starting_idx << endl << endl;
	cout << "=> Reading the unclassified-classified protein score file..." << endl;
	cout.flush();
		
	while( getline(fp_pair, lineStr) ){

		// Save information from the input file	
		tmp_pair[0].assign("");	// 1st column
		tmp_pair[1].assign("");	// 2nc column
		tmp_pair[2].assign("");	// 3rd column
		tmp_pair[3].assign("0.0"); // 4th column
		
		// cout << lineStr << endl;;
		istringstream iss;
		iss.str(lineStr);
		for(arr_n_idx = 0; arr_n_idx < 3; arr_n_idx++){
			iss >> lineStr_arr;
			tmp_pair[arr_n_idx].assign(lineStr_arr);
		}
		
		// Check seq overlapping percent here
		if( STRTOLD(tmp_pair[3]) < seq_overlapping_cutoff ){
			continue;
		}
		
		check_flag = 0;
		if( !PROTEIN_NAME_EXISTS(tmp_pair[0]) ){
			//cerr << "\nError : A protein '" << tmp_pair[0] << "' does not exist in the unclassified protein list file\n\n";
			check_flag = 1;
			//exit(0);
		}
		if( !PROTEIN_NAME_EXISTS(tmp_pair[1]) ){
			//cerr << "\nError : A protein '" << tmp_pair[1] << "' does not exists in the classified protein family list file\n\n";
			check_flag = 1;
			//exit(0);
		}
		
		if( check_flag == 1) continue;
		
		if( STRTOLD(tmp_pair[2]) == 0 ) {
			evalue = MAX_EVALUE;
		} else {
			evalue = -log(STRTOLD(tmp_pair[2]));
		}

		PUT_DIST_MAT(protein_name2id[tmp_pair[0]], protein_name2id[tmp_pair[1]], evalue, EVALUE_VAL);
		
		// Unknown protein list
		if( protein_name2id[tmp_pair[0]] != protein_name2id[tmp_pair[1]] ){
			unknown_protein_pair[protein_name2id[tmp_pair[0]]][unknown_protein_pair[protein_name2id[tmp_pair[0]]].size()] = protein_name2id[tmp_pair[1]];
		}

		// cout << tmp_pair[0] << "|\t" << tmp_pair[1] << "|\t" << tmp_pair[2] << "|\t" << tmp_pair[3] << "|\n";
	}

	fp_pair.close();
	cout  << endl;
	cout.flush();
}


void READ_TEST2TEST_SET_FILES(string test2test_pair_file_name, float seq_overlapping_cutoff)
{
	if( test2test_pair_file_name == "" ) return;
	
	ifstream fp_test2test_pair(test2test_pair_file_name.c_str());
	
	char ch, tmp_char[2], tmp_ID[20];
	string tmp_pair[4], lineStr, lineStr_arr;
	int i, j=0, tmp_i=0, tmp_score, tmp_idx, starting_idx = 0, arr_n_idx, check_flag = 0;
	float score_value, evalue;

	if( fp_test2test_pair.fail() ){
		cout << "Error!  Unable to open the unclassified-unclassified protein score file [" << test2test_pair_file_name << "] " << endl<< endl<< endl;
		exit(1);
	} 
	
	cout << "=> Reading the unclassified-unclassified protein score file..." << endl;
	cout.flush();
		
	while( getline(fp_test2test_pair, lineStr) ){

		// Save information from the input file	
		tmp_pair[0].assign("");	// 1st column
		tmp_pair[1].assign("");	// 2nc column
		tmp_pair[2].assign("");	// 3rd column
		tmp_pair[3].assign("0.0"); // 4th column
		
		// cout << lineStr << endl;;
		istringstream iss;
		iss.str(lineStr);
		for(arr_n_idx = 0; arr_n_idx < 3; arr_n_idx++){
			iss >> lineStr_arr;
			tmp_pair[arr_n_idx].assign(lineStr_arr);
		}
		
		// Save information from the input file
		if( STRTOLD(tmp_pair[3]) < seq_overlapping_cutoff ){ 		// Check seq overlapping percent here
			continue;
		}

		check_flag = 0;
		if( !PROTEIN_NAME_EXISTS(tmp_pair[0]) ){
			//cerr << "\nError : A protein '" << tmp_pair[0] << "' does not exist in the classified protein list file\n\n";
			check_flag = 1;
			//exit(0);
		}
		if( !PROTEIN_NAME_EXISTS(tmp_pair[1]) ){
			//cerr << "\nError : A protein '" << tmp_pair[1] << "' does not exists in the classified protein list file\n\n";
			check_flag = 1;
			//exit(0);
		}
		
		if( check_flag == 1) continue;
		
		if( STRTOLD(tmp_pair[2]) == 0 ) {
			evalue = MAX_EVALUE;
		} else {
			evalue = -log(STRTOLD(tmp_pair[2]));
		}

		PUT_TEST2TEST_DIST_MAT(protein_name2id[tmp_pair[0]], protein_name2id[tmp_pair[1]], evalue, EVALUE_VAL);
		//cout << "PUT[" << tmp_pair[0] << "] : " << evalue << "\tGET[" << tmp_pair[1] << "] : " << GET_TEST2TEST_DIST_MAT(protein_name2id[tmp_pair[0]], protein_name2id[tmp_pair[1]], EVALUE_VAL) << endl;
					
		// Unknown protein list
		if( protein_name2id[tmp_pair[0]] != protein_name2id[tmp_pair[1]] ){
			unknown_to_unknown_pair[protein_name2id[tmp_pair[0]]][unknown_to_unknown_pair[protein_name2id[tmp_pair[0]]].size()] = protein_name2id[tmp_pair[1]];
		}

		// cout << tmp_pair[0] << "|\t" << tmp_pair[1] << "|\t" << tmp_pair[2] << "|\t" << tmp_pair[3] << "|\n";
	}

	fp_test2test_pair.close();
	cout << endl;
	cout.flush();
}


float GET_AVG_HITS(int src_idx, string f_name)
{
	int i, trg_idx, cnt = 0;
	float sum = 0.0, tmp_val;

	for(i=0; i<family_id_list[f_name].size(); i++){	
		trg_idx = family_id_list[f_name][i];
		tmp_val = GET_DIST_MAT(src_idx, trg_idx, EVALUE_VAL);
		if( tmp_val != MIN_EVALUE ){
			sum += tmp_val;
			cnt++;
		}
	}
	
	return sum / (float) cnt;
}


map<int, int> GET_RAND_NUM(int num)
{
	map<int, int> return_val, real_return_val;
	map<int, int>::iterator iter;
	
	int i, rnd_num;
	
	for(i=0; i<num; i++){
		do{
			rnd_num = rand() % (num * 100);
		} while( return_val.find(rnd_num) != return_val.end() );
		return_val[rnd_num] = i;
	}
	
	i = 0;
	for(iter = return_val.begin(); iter != return_val.end(); iter++){
		real_return_val[i++] = (*iter).second;
	}
	
	return real_return_val;
}


float GET_AVG_FAM(map<int, int> u_set, float eval_cutoff, string *final_fam)
{
	map<int, int>::iterator iter;
	map<string, int> cand_fams, fam_avg_hit_counts;
	map<string, int>::iterator iter_cand_fams;
	map<string, float> fam_avg_hits;

	int i, j, p_id[2];
	float top_value, tmp_value;
	string top_family;

	for(iter = u_set.begin(); iter != u_set.end(); iter++)
	{
		// cout << (*iter).first << endl;
		p_id[0] = (*iter).first;
		if( p_id[0] == -1 ) continue;

		cand_fams.clear();
		for(j=0; j<unknown_protein_pair[p_id[0]].size(); j++)
		{
			p_id[1] = unknown_protein_pair[p_id[0]][j];
			if( protein_id2family[p_id[1]] == UNKNOWN_FAMILY){
				continue;
			}

			tmp_value = GET_DIST_MAT(p_id[0], p_id[1], EVALUE_VAL);
		
			if( eval_cutoff > tmp_value) {
				continue;
			}
			cand_fams[protein_id2family[p_id[1]]] = 1;
		}

		// cout << protein_id2name[p_id[0]] << ", "; 
		for( iter_cand_fams=cand_fams.begin(); iter_cand_fams != cand_fams.end(); iter_cand_fams++)
		{
			fam_avg_hit_counts[(*iter_cand_fams).first] += 1;
			fam_avg_hits[(*iter_cand_fams).first] += GET_AVG_HITS( p_id[0], (*iter_cand_fams).first );
		}
	}
	// cout << endl << "\t";

	top_value = -100;
	for( iter_cand_fams=fam_avg_hit_counts.begin(); iter_cand_fams != fam_avg_hit_counts.end(); iter_cand_fams++)
	{
		tmp_value = fam_avg_hits[(*iter_cand_fams).first] / (float)(fam_avg_hit_counts[(*iter_cand_fams).first]);
		// cout << (*iter_cand_fams).first << "[" << tmp_value << "] ";

		if( tmp_value > top_value ) {
			top_family.assign((*iter_cand_fams).first);
			top_value = tmp_value;
		}
	}
	
	*final_fam = top_family;

	return top_value;
}


float GET_AVG_UNKNOWN(map<int, map<int, int> > u_set, int cur_idx, int* selected_idx, float eval_cutoff)
{
	int i, top_idx = cur_idx, src_idx, trg_idx, idx_count;
	float top_value = MIN_EVALUE, tmp_value, tmp_value2;
	map<int, int>::iterator iter1, iter2;

	for( i = cur_idx + 1; i < u_set.size(); i++ )
	{
		tmp_value = MIN_EVALUE;
		idx_count = 0;
		for( iter1 = u_set[cur_idx].begin(); iter1 != u_set[cur_idx].end(); iter1++ )
		{
			src_idx = (*iter1).first;
			for( iter2 = u_set[i].begin(); iter2 != u_set[i].end(); iter2++ )
			{
				trg_idx = (*iter2).first;
				tmp_value2 = GET_TEST2TEST_DIST_MAT(src_idx, trg_idx, EVALUE_VAL);
				if( tmp_value2 != MIN_EVALUE ){
					tmp_value += tmp_value2;
					idx_count++;
				}
			}
		}
		
		if( idx_count == 0 ) continue;
		
		// cout << "\t" << protein_id2name[src_idx] << "," << protein_id2name[trg_idx] << " " << tmp_value << endl;
		
		tmp_value = tmp_value / (float) idx_count;
		if( tmp_value > top_value && tmp_value > eval_cutoff ){
			top_value = tmp_value;
			top_idx = i;
		}
	}
	
	*selected_idx = top_idx;

	return top_value;
}


void CLASSIFY(float eval_cutoff)
{
	map<int, int>::iterator iter;
	map<int, map<int, int> > u_set;
	int i, p_id, idx;
	string fam_cand;
	float fam_val, test_val;
	time_t cur_time;
	char tmp_top_family[200];
	
	for(i=0; i<family_id_list[UNKNOWN_FAMILY].size(); i++)
	{
		u_set[i][family_id_list[UNKNOWN_FAMILY][i]] = 1;
	}

	for(i=0; i<u_set.size(); i++)
	{
		idx = -1;
		// for(iter = u_set[i].begin(); iter != u_set[i].end(); iter++)
		// 	cout << "[" << (*iter).first << "]" << endl;
		// cout << endl;
		fam_cand.assign("");
		
		fam_val  = GET_AVG_FAM(u_set[i], eval_cutoff, &fam_cand);
		test_val = GET_AVG_UNKNOWN(u_set, i, &idx, eval_cutoff);
		// cout << "fam_val : " << fam_val << ", test_val : " << test_val << endl;
		
		if( fam_val >= test_val ){
			// ADD NEW FAMILY
			// cout << " - New fam [" << fam_cand << "] ";
			for( iter = u_set[i].begin(); iter != u_set[i].end(); iter++ )
			{
				p_id = (*iter).first;
				output_file_fp << protein_id2name[p_id] << "\t" << fam_cand << endl;

				// Assign a new family into unknow protein  and update std and avg of the family ids
				protein_id2family[p_id].assign(fam_cand);
				family_id_list[fam_cand][family_id_list[fam_cand].size()] = p_id;

			}

			
		} else {
			if( i == idx ){
				time(&cur_time);
				sprintf(tmp_top_family,"NewFamily_%d_%d",(int)cur_time,i);
				fam_cand.assign(tmp_top_family);
				for( iter = u_set[i].begin(); iter != u_set[i].end(); iter++ )
				{
					p_id = (*iter).first;
					output_file_fp << protein_id2name[p_id] << "\t" << fam_cand << endl;

					// Assign a new family into unknow protein  and update std and avg of the family ids
					protein_id2family[p_id].assign(fam_cand);
					family_id_list[fam_cand][family_id_list[fam_cand].size()] = p_id;
				}
			} else {

				// MOVE TO ANOTHER SET
				// cout << " - Move to another set from " << i << " to " << idx <<  "[" << test_val << "] = ";
				for( iter = u_set[i].begin(); iter != u_set[i].end(); iter++ )
				{
					u_set[idx][(*iter).first] = 1;
				}
			
				// for( iter = u_set[idx].begin(); iter != u_set[idx].end(); iter++ )
				// 	cout << protein_id2name[(*iter).first] << " ";
			}
		}
		// cout << "\n\t" << fam_cand[0] << "," << fam_cand[1] << endl;
	}
}


void INIT_VARS()
{
	// Data initialization 
	extern map<int, map<int, float> > distance_matrix_score;		// Distance matrix of blast score
	extern map<int, map<int, float> > distance_matrix_evalue;		// Distance matrix of blast e-value by -log(e-value)
	extern map<int, string> protein_id2name, protein_id2family;		// protein idx returns its protein name, protein_id2family returns its family name
	extern map<string, int> protein_name2id;				// protein name returns its protein idx.
	extern map<string, map<int, int> > family_id_list;			// family name returns its protein idxs. family_id_list[family_name][#] = [protein id]
	extern map<int, map<int, int> > unknown_protein_pair;			// unknown_proteins[protein_id][protein_id] = 1

}


int main(int argc, char* argv[])
{
	string training_family_file_name, training_pair_file_name, testing_family_file_name, testing_pair_file_name, out_file_name;
	string feature_set_file_name, test_test_pair_file_name;
	time_t start_time, end_time;
	int n_number;
	float eval_cutoff, seq_overlapping_cutoff = -1;
	
	time(&start_time);
	srand( (unsigned)time(NULL));
	
	CHECK_START(argc, argv, &training_family_file_name, &testing_family_file_name, &testing_pair_file_name, &test_test_pair_file_name, &eval_cutoff, &out_file_name);
	// INIT_VARS();

	cout << "=> e-value cutoff : " << eval_cutoff << endl << endl;
	eval_cutoff = -log(eval_cutoff);

	// Read input data
	READ_TRAINING_SET_FILES(training_family_file_name);
	READ_TESTING_SET_FILES(testing_family_file_name, testing_pair_file_name, seq_overlapping_cutoff);
	READ_TEST2TEST_SET_FILES(test_test_pair_file_name, seq_overlapping_cutoff);
	
	output_file_fp.open(out_file_name.c_str());
	// CHECK_UNKNOWN(eval_cutoff);
	CLASSIFY(eval_cutoff);
	output_file_fp.close();
	
	time(&end_time);
	ELAPSED_TIME(start_time, end_time);

	return 0;
}

